UNIT Board;

INTERFACE
USES
  Globals;

PROCEDURE SetUpStartPosition;
PROCEDURE PutPiece   (P : PieceType; Sq : Int8);    {-- places the piece on Square }
PROCEDURE RemovePiece(Sq : Int8);                   {-- sets the piece list to -1 and the board piece to 0 }
PROCEDURE ReadFen;
PROCEDURE DisplayTheBoard;
PROCEDURE WriteFen;
PROCEDURE SetUpBoard;


IMPLEMENTATION
USES
  Crt,
  Utils,
  Zorbist,
  TT,
  Debugging,
  Opening,
  Move;


{$Assertions On}

{---------------------------------------------------------------------------------------------}

PROCEDURE SetUpStartPosition;
VAR
  i : Byte;
BEGIN
  FillChar(Data, SizeOf(Data), 0);
  InitZorbistBoard;
  ZeroTTs;

  Ply := 0;
  WITH Data[Ply] DO
    BEGIN
      {-- place the pawns }
      Board[1] := %00010001000100010001000100010001;  {-- white pawns = 1 }
      Board[6] := %10011001100110011001100110011001;  {-- black pawns = 9 }

      {-- place the pieces }
{rank0}Board[0] := Wrook OR (Wnight SHL 4) OR (Wbishop SHL 8) OR (Wqueen SHL 12) OR (Wking SHL 16) OR
                  (Wbishop SHL 20) OR (Wnight SHL 24) OR (Wrook SHL 28);
{rank7}Board[7] := Brook OR (Bnight SHL 4) OR (Bbishop SHL 8) OR (Bqueen SHL 12) OR (Bking SHL 16) OR
                  (Bbishop SHL 20) OR (Bnight SHL 24) OR (Brook SHL 28);

      {-- init the piece lists }
      FOR i := 0 TO 7 DO
        BEGIN
          WPawnList[i] := i +  8;
          BPawnList[i] := i + 48;
        END;

      WPieceList[0, Wnight] := 1;
      WPieceList[1, Wnight] := 6;

      WPieceList[0, Wbishop] := 2;
      WPieceList[1, Wbishop] := 5;

      WPieceList[0, Wrook] := 0;
      WPieceList[1, Wrook] := 7;

      WPieceList[0, Wqueen] := 3;
      WPieceList[1, Wqueen] := NoPiece;

      WPieceList[0, Wking]  := 4;
      WPieceList[1, Wking]  := NoPiece;

      BPieceList[0, Bnight] := 57;
      BPieceList[1, Bnight] := 62;

      BPieceList[0, Bbishop] := 58;
      BPieceList[1, Bbishop] := 61;

      BPieceList[0, Brook] := 56;
      BPieceList[1, Brook] := 63;

      BPieceList[0, Bqueen] := 59;
      BPieceList[1, Bqueen] := NoPiece;

      BPieceList[0, Bking]  := 60;
      BPieceList[1, Bking]  := NoPiece;

      InitOpeningLib;

      CastleStat := %00001111;  {-- set castling flags }
      EnPasSq    := -1;
      WhosMove   := Wmove;
      MoveNum    := 0;
      HashKey    := CalcBrdKey;

      InitTakeBackStack;
    END; {-- with }
END;


PROCEDURE RemoveFromPieceList(P : PieceType; Sq : Int8);
VAR
  i : Byte;
BEGIN
  Assert(P IN [2..6, 10..14], 'Error in RemoveFromPieceList');

  WITH Data[Ply] DO
    BEGIN
      FOR i := 0 TO 1 DO     {-- two of each piece }
        IF P < Bpawn THEN    {-- white }
          BEGIN
            IF WpieceList[i, P] = Sq THEN
              WpieceList[i, P] := NoPiece;
          END
        ELSE  {-- black }
          IF BpieceList[i, P] = Sq THEN
            BpieceList[i, P] := NoPiece;
    END; {-- for }
END;


PROCEDURE AddToPieceList(P : PieceType; Sq : Int8);
BEGIN
  Assert(P IN [2..9, 10..14],'Error in AddToPieceList');

  WITH Data[Ply] DO
    BEGIN
      IF P < Bpawn THEN  {-- white }
        BEGIN
          IF WpieceList[0, P] = NoPiece THEN
            WpieceList[0, P] := Sq
          ELSE
            WpieceList[1, P] := Sq;
        END
      ELSE  {-- black }
        IF BpieceList[0, P] = NoPiece THEN
          BpieceList[0, P] := Sq
        ELSE
          BpieceList[1, P] := Sq;
    END;  {-- with }
END;


PROCEDURE RemoveFromPawnList(Sq : Int8);
{-- scan for the sq and set it to -1 }
VAR
  i : Int8;
  P : PieceType;
BEGIN
  P := WhatPiece(Sq);
  IF WhatPiece(Sq) = Empty THEN  {-- it will be for an Enpasaunt capture }
    Exit;

  Assert(P IN [Wpawn, Bpawn], 'error: in removefrompawnlist');

  WITH Data[Ply] DO
    BEGIN
      IF P = Wpawn THEN
        BEGIN
          i := -1;
          REPEAT
            Inc(i);
            Assert(i < 8, 'Error in AddToPawnList, i value > 7, white code');
          UNTIL WPawnList[i] = Sq;   {-- found the piece }
          WPawnList[i] := NoPiece;
        END;

      IF P = Bpawn THEN
        BEGIN
          i := -1;
          REPEAT
            Inc(i);
            Assert(i < 8, 'Error in AddToPawnList, i value > 7, white code');
          UNTIL BPawnList[i] = Sq;
          BPawnList[i] := NoPiece;
        END;
  END; {-- with }
END;


PROCEDURE AddToPawnList(Sq : Int8);
{-- look for an empty slot and place the sq there }
VAR
  i : Int8;
  P : PieceType;
BEGIN
  P := WhatPiece(Sq);
  Assert(P IN [Wpawn, Bpawn], 'error: in AddToPawnList');

  WITH Data[Ply] DO
    BEGIN
      i := -1;
      IF P = Wpawn THEN
        BEGIN
          REPEAT  {-- find the first empty slot }
            Inc(i);
            Assert(i < 8, 'Error in AddToPawnList, i value > 7, white code');
          UNTIL (WpawnList[i] = NoPiece);
          WpawnList[i] := Sq;
        END
      ELSE
        BEGIN
          REPEAT
            Inc(i);
            Assert(i < 8, 'Error in AddToPawnList, i value > 7, black code');
          UNTIL (BpawnList[i] = NoPiece);
          BpawnList[i] := Sq;
        END;
  END;  {-- with }
END;


PROCEDURE PutPiece({piece type}P  : PieceType;
                   {onto}      Sq : Int8);
CONST
  {-- mask to isolate a square }
  Mask0 = %11111111111111111111111111110000;
  Mask1 = %11111111111111111111111100001111;
  Mask2 = %11111111111111111111000011111111;
  Mask3 = %11111111111111110000111111111111;
  Mask4 = %11111111111100001111111111111111;
  Mask5 = %11111111000011111111111111111111;
  Mask6 = %11110000111111111111111111111111;
  Mask7 = %00001111111111111111111111111111;

VAR
  Rank,
  Fyle : Byte;
  Temp : Cardinal;
BEGIN
  Rank := Sq SHR 3;
  Fyle := Sq MOD 8;

  {-- zero the nibble/square, cause we may be putting nopiece(clearing a square) }
  WITH Data[Ply] DO
    CASE Fyle OF
      0 : Board[Rank] := Board[Rank] AND Mask0;
      1 : Board[Rank] := Board[Rank] AND Mask1;
      2 : Board[Rank] := Board[Rank] AND Mask2;
      3 : Board[Rank] := Board[Rank] AND Mask3;
      4 : Board[Rank] := Board[Rank] AND Mask4;
      5 : Board[Rank] := Board[Rank] AND Mask5;
      6 : Board[Rank] := Board[Rank] AND Mask6;
      7 : Board[Rank] := Board[Rank] AND Mask7;
    END;

  {-- if placing NoPiece then nothing more to do }
  IF (P <> Empty) THEN
    BEGIN
      Temp := Cardinal(P) SHL (4*Fyle);  {-- move to the correct nibble }

      {-- place the piece }
      Data[Ply].Board[Rank] := Data[Ply].Board[Rank] OR Temp;

      {-- update zorbist key }
      Data[Ply].HashKey := Data[Ply].HashKey XOR ZorbistBrd[P, Sq];

      {-- update the piece lists }
      IF P IN [Wpawn, Bpawn] THEN
        AddToPawnList(Sq)
      ELSE
        AddToPieceList(P, Sq);
    END;
END;


PROCEDURE RemovePiece(Sq : Int8);
VAR
  P : PieceType;
BEGIN
  P := WhatPiece(Sq);

  IF P <> Empty THEN  {-- it is for enpassant }
    BEGIN
      {-- remove the piece from the piece lists }
      IF P IN [Wpawn, Bpawn] THEN
        RemoveFromPawnList(Sq)
      ELSE
        RemoveFromPieceList(P, Sq);

      {-- updata zorbist key }
      Data[Ply].HashKey := Data[Ply].HashKey XOR ZorbistBrd[P, Sq];

      {-- replace the piece with 0 }
      PutPiece(Empty, Sq);
    END;
END;


PROCEDURE ZeroBoardData;
BEGIN
  WITH Data[Ply] DO
    BEGIN
      FillChar(Board,      SizeOf(BoardType), 0);
      FillChar(WpieceList, SizeOf(WpieceList), Byte(-1));
      FillChar(BpieceList, SizeOf(BpieceList), Byte(-1));
      FillChar(WpawnList,  SizeOf(WpawnList),  Byte(-1));
      FillChar(BpawnList,  SizeOf(BpawnList),  Byte(-1));
      Data[Ply].CastleStat := 0;
    END;
END;


PROCEDURE ReadFen;
{-- reads the board position from file in almost FEN format }
VAR
  f     : Text;       {-- board position }
  Ch    : Char;
  S     : String[10];
  Piece : PieceType;
  Rank,
  Fyle,
  Sq    : Int8;
BEGIN

  Assign(f, 'game.fen');
{$I-}  Reset(f);  {$I+}

  IF IOResult = 0 THEN
    BEGIN
      ZeroBoardData;

      Rank := 8;
      REPEAT
        Dec(Rank);
        Fyle := 0;
        REPEAT
          Read(f, Ch);
          Piece := NoPiece;

          IF Ch IN ['1'..'8'] THEN
            Inc(Fyle, (Ord(Ch) - 48) - 1)  {-- increments 0..7 }
          ELSE
            CASE Ch OF
             'R' : Piece := Wrook;
             'N' : Piece := Wnight;
             'B' : Piece := Wbishop;
             'Q' : Piece := Wqueen;
             'K' : Piece := Wking;
             'P' : Piece := Wpawn;
             'r' : Piece := Brook;
             'n' : Piece := Bnight;
             'b' : Piece := Bbishop;
             'q' : Piece := Bqueen;
             'k' : Piece := Bking;
             'p' : Piece := Bpawn;
            END; {case}

          Sq := 8*Rank + Fyle;
          IF Piece <> NoPiece THEN
            PutPiece(Piece, Sq);  {-- onto the board }

          Inc(Fyle);
        UNTIL Fyle = 8;

        Read(f, Ch);         {-- '/' char }
      UNTIL Rank = 0;

      ReadLn(f, S);

      IF (Pos('w', S)  <> 0) THEN
        WhosMove := 1
      ELSE
        WhosMove := -1;

      {-- set castling rights }
      WITH Data[Ply] DO
        BEGIN
          CastleStat := 0;
          IF Pos('K', S) <> 0 THEN
            CastleStat := %00000001;
          IF Pos('Q', S) <> 0 THEN
            CastleStat := CastleStat OR %00000010;

          IF Pos('k', S) <> 0 THEN
            CastleStat := CastleStat OR %00000100;
          IF Pos('q', S) <> 0 THEN
            CastleStat := CastleStat OR %00001000;
        END;
{
        IF Pos(' cw', S) <> 0 THEN
          ComputerPlays := WhitePieces
        ELSE
          ComputerPlays := BlackPieces;

        Enpasauntsq := NoPiece;
}
      Close(f);

      Data[Ply].EnPasSq := -1;
      ZeroTTs;
      InitZorbistBoard;

      InLibrary := False;
    END {-- if }
  ELSE
    ErrorMsg('No Board File Found');
END;


PROCEDURE WriteFen;
{-- writes the board position to file in almost FEN format }
CONST
  PieceChar : ARRAY[1..14] OF Char = 'PNBRQK**pnbrqk';
VAR
  f       : Text;                     {-- for board position }
  Files,
  Ranks,
  Square,
  Vacant,
  Piece  : Byte;
BEGIN
  Assign(f, 'game.fen');
  Rewrite(f);
  FOR Ranks := 7 DOWNTO 0 DO
    BEGIN
      Vacant := 0;
      FOR Files := 0 TO 7 DO
        BEGIN
          Square := 8*Ranks + Files;
          Piece := WhatPiece(Square);
          IF Piece = Empty THEN
            Inc(Vacant)
          ELSE
            BEGIN
              IF Vacant <> 0 THEN
                BEGIN
                  Write(f, Char(Vacant + 48));
                  Vacant := 0;
                END;
              Write(f, PieceChar[Piece]);
            END;
          IF (Vacant <> 0) AND (Files = 7) THEN
            Write(f, Char(Vacant + 48));
        END;
      IF Ranks > 0 THEN
        Write(f, '/');
    END;

  IF WhosMove = 1 THEN
    Write(f, ' w ')
  ELSE
    Write(f, ' b ');

  WITH Data[Ply] DO
    BEGIN
      IF (CastleStat AND %00000001) = 1 THEN
        Write(f, 'K');
      IF (CastleStat AND %00000010) = 2 THEN
        Write(f, 'Q');
      IF (CastleStat AND %00000100) = 4 THEN
        Write(f, 'k');
      IF (CastleStat AND %00001000) = 8 THEN
        Write(f, 'q');
    END;
(*
  IF NOT HumansMove THEN
    Write(f, ' cm');

  IF ComputerPlays = WhitePieces THEN
    Write(f, ' cw')
  ELSE
    Write(f, ' cb');
  *)
  Close(f);
END;


PROCEDURE DisplayTheBoard;
CONST
  PieceChar : String[14] = 'PNBRQK  pnbrqk';
VAR
  Sq    : Int8;
  Rank  : Byte;
  Swap  : Boolean;
  Piece : PieceType;
BEGIN
  HighVideo;
  TextBackGround(Black);
  Window(1, 1, 27, 10);
  ClrScr;
  GoToXY(1, 2);

  Swap := False;
  Sq   := 56;
  Rank := 9;

  REPEAT      {-- for each Rank }
    Dec(Rank);
    TextBackGround(Black);
    TextColor(White);
    Write(Rank);

     FOR Sq := Sq TO (Sq + 7) DO   {-- print file 8 down to 1 }
      BEGIN
        TextColor(White);

        IF NOT Swap THEN
          TextBackGround(Green)
        ELSE
          TextBackGround(Yellow);

        Piece := WhatPiece(Sq);
        IF Piece IN [Wpawn..Wking] THEN    {-- white pieces }
          TextColor(White)
        ELSE
          TextColor(Black);

        IF Piece = Empty THEN
          Write('   ')
        ELSE  {-- index the char in piecechar }
          Write(PieceChar[Ord(Piece)] : 2, ' ');

        Swap := NOT Swap;
      END;  {-- for }


    Writeln;
    Swap := NOT Swap;
    Dec(Sq, 15);
  UNTIL Sq = -8;

  TextBackGround(Black);
  TextColor(White);

  Write  ('  a  b  c  d  e  f  g  h');

  {-- now write a menu }
  Window(30, 2, 48, 13);
  ClrScr;
  Writeln('n: New Game');
  Writeln('b: TakeBack');
  Writeln('p: Play    ');
  Writeln('s: Setup   ');
  Writeln('w: Write game');
  Writeln('r: Read game');
  Writeln('t: Move Time(tx)');
  WriteLn('x: Fixed Ply(xx)');
  Write  ('q: Quit');
  Window(1, 1, 80, 25);

  GoToXY(6, 11);
  IF FixedDepth < 255 THEN
    Write('Fixed Depth = ', FixedDepth, '     ')
  ELSE
    Write('Move Time = ', MoveTime, ' secs    ');

  LowVideo;
END;


PROCEDURE SetUpBoard;
{-- to place a piece...'wbd4' = white bishop on D4
    to clear a square..'e5' = clear square E5
    to set the colour for the next move...'wm' or 'bm' (WhiteMove or BlackMove
    to quit.... 'q' }
VAR
  i        : 1..4;
  InStr    : String[4];
  Piece    : PieceType;
  Sq       : Int8;
  DelPiece : Boolean;


  PROCEDURE WriteSetupMenu;
  BEGIN
    GoToXY(1, 15);
    Writeln(' c:    clear board       ');
    Writeln(' wnf3: white knight on f3');
    Writeln(' g1:   clear square g1   ');
    Writeln(' wm:   white to move     ');
    Writeln(' q:    quit              ');
  END;


BEGIN
  ClrScr;

  WITH Data[Ply] DO
    REPEAT
      DisplayTheBoard;
      WriteSetupMenu;
      DelPiece := False;
      Piece := NoPiece;

      {-- get the user  input }
      GoToXY(31, 12);
      Write('Command:                    ');
      GoToXY(40, 12);
      Readln(InStr);

      FOR i := 1 TO Length(InStr) DO
        InStr[i] := UpCase(InStr[i]);

      {-- clear board }
      IF InStr = 'C' THEN
        BEGIN
          FillChar(Data[Ply].Board, SizeOf(BoardType), 0);
          FillChar(Data[Ply].WPawnList, SizeOf(PawnListType), Byte(-1));
          FillChar(Data[Ply].BPawnList, SizeOf(PawnListType), Byte(-1));
          Continue;
        END;

      IF Length(InStr) = 2 THEN
        BEGIN
          IF Instr = 'WM' THEN
            WhosMove := 1
          ELSE
            IF Instr = 'BM' THEN
              WhosMove := -1
            ELSE
              IF (InStr[1] IN ['A'..'H']) AND (InStr[2] IN ['1'..'8']) THEN
                DelPiece := True
              ELSE
                IF (Instr <> 'WM') AND (Instr <> 'BM') THEN
                  Instr := '__' + Instr;
        END;

      {-- get the piece type and colour }
      IF Length(Instr) = 4 THEN
        BEGIN
          CASE InStr[1] OF
            'W' : CASE InStr[2] OF
                    'P' : Piece := WPawn;
                    'R' : Piece := WRook;
                    'N' : Piece := Wnight;
                    'B' : Piece := WBishop;
                    'Q' : Piece := WQueen;
                    'K' : Piece := WKing;
                  END; {-- case }
            'B' : CASE InStr[2] OF
                    'P' : Piece := BPawn;
                    'R' : Piece := BRook;
                    'N' : Piece := Bnight;
                    'B' : Piece := BBishop;
                    'Q' : Piece := BQueen;
                    'K' : Piece := BKing;
                  END; {-- case }
          END; {-- case }
        END;   {-- if }

      {-- insert or remove a piece }
      IF Length(InStr) = 4 THEN
        Sq := StrToToSq(InStr);    {-- reusing these, but are badly named here }
      IF Length(InStr) = 2 THEN
        Sq := StrToFromSq(InStr);  {-- reusing these, but are badly named here }

      IF InStr <> 'Q' THEN
        IF ( (Piece = WPawn) AND (Sq < 8) ) OR
           ( (Piece = BPawn) AND (Sq > 55) ) THEN
           ErrorMsg('Illegal position')
        ELSE  {-- legal placement }
          IF Sq IN [0..63] THEN
            IF DelPiece THEN
              RemovePiece(Sq)
            ELSE  {-- add piece }
              PutPiece(Piece, Sq);

    UNTIL InStr = 'Q';

  MoveNum := 0;
  HumansMove := True;

  ClrScr;
  InitZorbistBoard;
  Data[Ply].HashKey := CalcBrdKey;
  ZeroTTs;

  InLibrary := False;
END;


BEGIN
END.


