UNIT Eval;

INTERFACE
USES
  Globals;


FUNCTION EvalPosition(Alpha, Beta : Int16) : Int16;


IMPLEMENTATION
USES
  LookUpTables,
  Utils;


CONST
  DoublePawnPenalty = 20;   {-- per pawn }
  NotCastPenalty    = 30;
  TwoBishopBonus    = 30;
  KingSafety        =  6;   {-- for each element of safety }


FUNCTION EvalPosition(Alpha, Beta : Int16) : Int16;
CONST
  CutOff = 300;
VAR
  LazyEval,
  Material,
  Score     : Int16;
  i         : Byte;
  PawnFiles : ARRAY[0..7] OF Byte;
BEGIN
  Material := CalcMaterial;

  {-- try lazy eval first  }
  IF NOT EndGame THEN
    BEGIN
      LazyEval := Material * WhosMove;
      IF (LazyEval + CutOff) < Alpha THEN
        Exit(Alpha);
      IF (LazyEval - CutOff) > Beta THEN
        Exit(Beta);
    END;

  {-- piece values on squares via PST's }
  Score := 0;
  WITH Data[Ply] DO
    BEGIN
      {-- pawns }
      FOR i := 0 TO 7 DO
        BEGIN
          {-- white }
          Inc(Score, Pst_Pawn[WpawnList[i]]);

          {-- black }
          Dec(Score, Pst_Pawn[Flip(BpawnList[i])]);
        END;

      FOR i := 0 TO 1 DO
        BEGIN
          {-- nights }
          Inc(Score, Pst_knight[     WpieceList[i, Wnight]]);
          Dec(Score, Pst_knight[Flip(BpieceList[i, Bnight])]);

          {-- bishop }
          Inc(Score, Pst_bishop[     WpieceList[i, Wbishop]]);
          Dec(Score, Pst_bishop[Flip(BpieceList[i, Bbishop])]);

          {-- rooks }
          Inc(Score, Pst_rook[     WpieceList[i, Wrook]]);
          Dec(Score,  Pst_rook[Flip(BpieceList[i, Brook])]);

          {-- queen }
          Inc(Score, Pst_queen[     WpieceList[i, Wqueen]]);
          Dec(Score, Pst_queen[Flip(BpieceList[i, Bqueen])]);

          IF NOT EndGame THEN
            BEGIN
              Inc(Score, Pst_king_mg[     WpieceList[0, Wking]]);
              Dec(Score, Pst_king_mg[Flip(BpieceList[0, Bking])]);
            END
          ELSE
            BEGIN
              Inc(Score, Pst_king_eg[     WpieceList[0, Wking]]);
              Dec(Score, Pst_king_eg[Flip(BpieceList[0, Bking])]);
            END;
        END;  {-- for }
    END;  {-- with }

  {-- just a few spices }
  WITH Data[Ply] DO
    BEGIN
      IF CastleStat AND %00000011 = %00000011 THEN  {-- white not castled }
        Dec(Score, NotCastPenalty);
      IF CastleStat AND %00001100 = %00001100 THEN  {-- white not castled }
        Inc(Score, NotCastPenalty);

      {-- two bishops }
      IF (WpieceList[0, Wbishop] <> NoPiece) AND (WpieceList[1, Wbishop] <> NoPiece) THEN
        Inc(Score, TwoBishopBonus);
      IF (BpieceList[0, Bbishop] <> NoPiece) AND (BpieceList[1, Bbishop] <> NoPiece) THEN
        Dec(Score, TwoBishopBonus);

      {-- king safety for white }
      IF CastleStat AND %00000011 = %00000010 THEN  {-- king side }
        BEGIN
          CASE WhatPiece(13) OF
            Wpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(14) OF
            Wpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(15) OF
            Wpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(21) OF
            Wnight : Inc(Score, KingSafety);
          END;
        END;
      IF CastleStat AND %00000011 = %00000001 THEN  {-- queen side }
        BEGIN
          CASE WhatPiece(8) OF
            Wpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(9) OF
            Wpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(10) OF
            Wpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(18) OF
            Wnight : Inc(Score, KingSafety);
          END;
        END;

      {-- king safety for black }
      IF CastleStat AND %00001100 = %00000100 THEN  {-- king side }
        BEGIN
          CASE WhatPiece(53) OF
            Bpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(54) OF
            Bpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(55) OF
            Bpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(45) OF
            Bnight : Inc(Score, KingSafety);
          END;
        END;
      IF CastleStat AND %00001100 = %00001000 THEN  {-- queen side }
        BEGIN
          CASE WhatPiece(48) OF
            Bpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(49) OF
            Bpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(50) OF
            Bpawn : Inc(Score, KingSafety);
          END;
          CASE WhatPiece(42) OF
            Wnight : Inc(Score, KingSafety);
          END;
        END;

      {-- pawn structure...only doing doubled pawns at the moment }
      FillChar(PawnFiles, SizeOf(PawnFiles), 0);

      {-- count the white pawns on each file }
      FOR i := 0 TO 7 DO
        IF WpawnList[i] <> -1 THEN
          Inc(PawnFiles[WhatFile(WpawnList[i])]);
      FOR i := 0 TO 7 DO
        IF PawnFiles[i] <> 0 THEN
          Score := Score - (PawnFiles[i] - 1)*DoublePawnPenalty;

      {-- count the black pawns on each file }
      FillChar(PawnFiles, SizeOf(PawnFiles), 0);

      FOR i := 0 TO 7 DO
        IF BpawnList[i] <> -1 THEN
          Inc(PawnFiles[WhatFile(BpawnList[i])]);
      FOR i := 0 TO 7 DO
        IF PawnFiles[i] <> 0 THEN
          Score := Score + (PawnFiles[i] - 1)*DoublePawnPenalty;

    END; {-- with }

  EvalPosition := WhosMove*(Score + Material);
END;




BEGIN
END.
