UNIT Search;



INTERFACE
USES
  Globals;


TYPE
  SaveNodesType = ARRAY[0..MaxPly] OF Word64; {-- used in calc' EBF }

VAR
  PV     : PVtype;
  Nodes,
  Qnodes : Word64;


FUNCTION ComputerMove : MoveRec;

{-----------------------------------------------------------------}
IMPLEMENTATION
USES
  Crt,
  Utils,
  GenMoves,
  Move,
  TT,
  Zorbist,
  Clocks,
  Eval,
  Debugging;


CONST
  MateScore = 10000;

VAR
  PVscore,
  CurrentDepth : Int16;  {-- of the iterative deepening search }

  {-- vars used for accounting }
  SaveNodes : SaveNodesType;
  Time,
  BetaCuts,
  NullCuts,
  TThits,
  ExactTThits : Word32;
  Abort       : Boolean;



PROCEDURE UpDatePV(Move : Str4; Depth : Byte);
VAR
  i : Byte;
BEGIN
  PV[Depth, Depth] := Move;
  FOR i := 1 TO CurrentDepth DO
    PV[Depth, Depth+i] := PV[Depth+1, Depth+i];
END;


FUNCTION TimesUp : Boolean;
BEGIN
  TimesUp := Clock(Elapsed) > MoveTime;
END;


FUNCTION QSearch(Alpha, Beta : Int16) : Int16;
VAR
  Score     : Int16;
  NumMoves,
  MoveIndex : Byte;
  ThisMove  : MoveRec;
  TTentry   : FullTTtype;
BEGIN
(*  {-- look for a TT hit first }
  TTentry.Key := Data[Ply].HashKey;
  GetFromTT(TTentry);
  IF TTentry.Node = Exact THEN
    Exit(TTentry.Score)
  ELSE
    IF (TTentry.Node = FailHigh) AND (TTentry.Score >= Beta) THEN
      Exit(Beta)
    ELSE    *)
      Score := EvalPosition(Alpha, Beta);

  IF Score > Alpha THEN
    Alpha := Score;

  IF Score >= Beta THEN
    BEGIN
      Inc(BetaCuts);
      Exit(Beta);
    END;

  IF Ply = MaxPly THEN
    Exit(Score);

  NumMoves := GenCaptureList;
  FOR MoveIndex := 1 TO NumMoves DO
    BEGIN
      ThisMove := GetAMove(NumMoves);
(*
      {-- poor mans SEE....but this worsens time to depth }
      IF (PieceValue[WhatPiece(ThisMove.ToSq)] < PieceValue[WhatPiece(ThisMove.FromSq)]) AND {-- high takes low }
          Attacked(ThisMove.ToSq, -1*WhosMove) THEN                                          {-- so if there is a recapture }
        Continue;                                                                            {-- ignore this move }
*)
      {-- move to next ply and copy the board data across }
      Inc(Ply);
      Data[Ply] := Data[Ply - 1];

      MakeMove(ThisMove, True);
      Score := -QSearch(-Beta, -Alpha);
      UnMakeMove;

      {-- now tally the score }
      IF Score > Alpha THEN
        BEGIN
          Alpha := Score;
          IF Alpha >= Beta THEN
            BEGIN
              Inc(BetaCuts);
              Exit(Beta);
            END;
        END;
    END;  {-- for }

  QSearch := Alpha;
END;


FUNCTION Search(Alpha, Beta : Int16; Depth : Int8; NullAllowed : Boolean) : Int16;
VAR
  NoMoves,
  FailLowNode,
  Check       : Boolean;
  NumMoves,
  MoveIndex   : Byte;
  Score       : Int16;
  ThisMove    : MoveRec;
  TTentry     : FullTTtype;
  TTmove      : Str4;
  R           : Byte;
BEGIN
  IF Abort THEN
    Exit(0);         {-- 'p' has been hit }

  Abort := ( (Nodes AND 8192) = 0 ) AND TimesUp;

  IF Abort THEN
    Exit(0);         {-- out of time }

  IF Ply = MaxPly THEN
    Exit(EvalPosition(Alpha, Beta));

  IF (Depth < 1) THEN
    Exit(Qsearch(Alpha, Beta));

  Check := InCheck(WhosMove);
  IF Check THEN
    Inc(Depth);

  {---------- look at the transposition table ---------------------}

  TTmove := 'A1A1';                  {-- sentinal value }
  TTentry.Key := Data[Ply].HashKey;
  GetFromTT(TTentry);

  IF ValidMove(TTentry) THEN         {-- usable info }
    BEGIN
      Inc(TThits);
      TTmove := TTentry.Move;        {-- will try this move first }

      IF TTentry.Depth >= Depth THEN
        BEGIN
          IF (TTentry.Node = FailHigh) AND (TTentry.Score >= Beta) THEN
            BEGIN
              Inc(BetaCuts);
              Exit(Beta);
            END;

          IF (TTentry.Node = FailLow) AND (TTentry.Score <= Alpha) THEN
            Exit(Alpha);

         IF TTentry.Node = Exact THEN
           BEGIN
             Inc(ExactTThits);
             Exit(TTentry.Score);
           END;  {-- if }
        END;
    END;


  {---------- try a null move looking for an early cutoff -----------------}

  IF NullAllowed AND
     NOT Check   AND
     NOT EndGame AND
     (Depth > 2) THEN    {-- not really worried about Zugzwang at this point }
    BEGIN
      SwapSides;
      Score := -Search(-Beta, -Beta + 1, Depth - 2, False);  {-- shallower search around Beta }
      SwapSides;

      IF (Score >= Beta) THEN
        BEGIN
          Inc(BetaCuts);
          Inc(NullCuts);
          AddToTT(Data[Ply].HashKey, Score, FailHigh, Depth, '', Empty);
          Exit(Beta);
        END;
    END;

  {------------------------ now search all moves ----------------------------}

  NoMoves     := True;   {-- unless proven otherwise }
  FailLowNode := True;   {-- unless proven otherwise }

  NumMoves := GenMoveList(TTmove);
  FOR MoveIndex := 1 TO NumMoves DO
    BEGIN
      ThisMove := GetAMove(NumMoves);

      {-- move to next ply and copy the board data across }
      Inc(Ply);
      Data[Ply] := Data[Ply - 1];

      MakeMove(ThisMove, False);

      IF InCheck(-1*WhosMove) THEN  {-- if you are in check then the move MUST get you out of check }
        BEGIN
          UnMakeMove;
          Continue;
        END
      ELSE
        NoMoves := False;

      {-- now the recursive call }
      R := 2 + Ord(Depth > 6);                                  {-- R = 2 and then 3 }
      IF (MoveIndex = 1) THEN                                   {-- PVS }
        Score := -Search(-Beta, -Alpha, Depth-1, True)          {-- full width window }
      ELSE
        BEGIN
          Score := -Search(-(Alpha+1), -Alpha, Depth-R, True);  {-- does it improve on Alpha }

          IF (Score > Alpha) AND (Score < Beta) THEN            {-- Doh! research }
            Score := -Search(-Beta, -Alpha, Depth-1, True);
        END;

      UnMakeMove;

      {-- now tally up the score }
      IF (Score > Alpha) THEN  {-- a better move }
        BEGIN
          Alpha       := Score;
          FailLowNode := False;

          IF (Alpha >= Beta) THEN   {-- Beta Cut }
            BEGIN
              Inc(BetaCuts);
              IF NOT ThisMove.Capture THEN
                AddToHistory(ThisMove, Depth);

              AddToTT(Data[Ply].HashKey, Score, FailHigh, Depth,
                      SquaresToStr(ThisMove.FromSq, ThisMove.ToSq),
                      WhatPiece(ThisMove.FromSq));
              Exit(Beta);
            END; {-- of beta cut }

          {-- score is > alpha and < beta }
          AddToTT(Data[Ply].HashKey, Score, Exact, Depth,
                SquaresToStr(ThisMove.FromSq, ThisMove.ToSq),
                WhatPiece(ThisMove.FromSq));

          UpDatePV(SquaresToStr(ThisMove.FromSq, ThisMove.ToSq), Ply);
        END;  {-- scoring }

      IF FailLowNode AND NOT NoMoves THEN
        AddToTT(Data[Ply].HashKey, Score, FailLow, Depth, '', Empty);   {-- there is no good move to store }

      {-- has the operator pressed the p key }
      IF (KeyPressed AND (ReadKey = 'p')) OR TimesUp THEN
        BEGIN
          Abort := True;
          Exit(Alpha);        {-- if times up the best move will be in the PV }
        END;

    END;  {-- all moves finished }

  {----------------- end of the move loop --------------------------------------------}

  IF NoMoves THEN
    IF NOT InCheck(WhosMove) THEN  {-- stalemate }
      Search := 0
    ELSE
      Search := -MateScore + Ply
  ELSE
    Search := Alpha;
END;


PROCEDURE InitSearch;
BEGIN
  GoToXY(49, 13);
  Write('I''m Thinking.......');

  {-- zero tables }
  FillChar(PV, SizeOf(PV), 0);
  FillChar(HistTab, SizeOf(HistTab), 0);
  ZeroTTs;

  {-- counters }
  Nodes       :=  0;  {-- to avoid divide by zero }
  QNodes      :=  0;
  BetaCuts    :=  0;
  NullCuts    :=  0;
  TThits      :=  0;
  ExactTThits :=  0;
  PVscore     :=  0;

  Abort    := False;           {-- abort ply if 'p' pressed }
  Time     := Clock(Start);
  CurrentDepth := 3;

  EndGame := IsEndGame;
END;


FUNCTION ComputerMove : MoveRec;
BEGIN
  InitSearch;

  WHILE NOT TimesUp AND                                 {-- for time per move }
        NOT Abort   AND                                 {-- 'p' key has been pressed }
        (CurrentDepth < FixedDepth) AND                 {-- for fixed ply per move }
        (Abs(PVscore) < (MateScore - 20)) DO            {-- forced mate found }
    BEGIN                                               {-- the iterative deepen loop }
      Inc(CurrentDepth);                                {-- search the next ply...iterative deepening }

      PVscore := Search(-30000, 30000, CurrentDepth, True); {-- call the search }

      Time := Clock(Elapsed);                           {-- check the time used so far }

      IF CurrentDepth > 2 THEN
        BEGIN
          SaveNodes[CurrentDepth] := Nodes + Qnodes - SaveNodes[CurrentDepth - 1];  {-- the nodes for each ply searched }
          IF Abs(PVscore) <> MateScore THEN
            WriteInfoToScreen(Nodes, Qnodes, BetaCuts, NullCuts, TThits, ExactTThits, PVscore, PV, EBF(SaveNodes, CurrentDepth), Time, CurrentDepth);
        END;
    END;  {-- of iterative deepening }

  {-- check that we haven't stuffed up during the search }
  Assert((Data[Ply].HashKey = CalcBrdKey), 'error zorbist key');

  IF Abs(PVscore) = MateScore THEN
    ComputerMove := MoveStrToMoveRec('MATE')
  ELSE
    ComputerMove := MoveStrToMoveRec(PV[0, 0]);  {-- first move at the top of the PV }
END;


BEGIN
END.
