UNIT Utils;


INTERFACE
USES
  SysUtils,
  Globals,
  TT,
  Search;



FUNCTION WhatPiece       (Sq : Int8)           : PieceType; INLINE;      {-- what piece is on this square }
FUNCTION WhatFile        (Sq : Int8)           : Byte; INLINE;
FUNCTION WhatRank        (Sq : Int8)           : Byte; INLINE;
FUNCTION StrToToSq       (Move : ShortStr)     : Byte; INLINE;
FUNCTION StrToFromSq     (Move : ShortStr)     : Byte; INLINE;
FUNCTION SquaresToStr    (FromSq, ToSq : Int8) : ShortStr; INLINE;
FUNCTION Attacked        (MySq, Enemy : Int8)  : Boolean;
FUNCTION InCheck         (Who  : Int8)         : Boolean;
FUNCTION ValidMove       (M : FullTTtype)      : Boolean;
FUNCTION MoveStrToMoveRec(Mstr : ShortStr)     : MoveRec; INLINE;
FUNCTION UpperCase       (S : String)          : String;
FUNCTION EBF(Var Nodes : SaveNodesType; Len : Byte) : Real;
FUNCTION IsMoveCmd(S : Str4) : Boolean;
FUNCTION Flip(Sq : Int8) : Int8; INLINE;
FUNCTION CalcMaterial : Int16;
FUNCTION IsEndGame : Boolean;

PROCEDURE AddToHistory(TheMove : MoveRec; Depth : Int8);
PROCEDURE SwapSides; INLINE;
PROCEDURE ErrorMsg(S : String);
PROCEDURE WritePV (VAR PV : PvType; Len : Byte);
PROCEDURE SetTime (S : Str4);
PROCEDURE SetDepth(S : Str4);
PROCEDURE WriteInfoToScreen(Nodes,
                            Qnodes,
                            BetaCuts,
                            NullCuts,
                            TThits,
                            ExactTThits : Word32;
                            PVscore     : Int16;
                            PV          : PvType;
                            EBF : Real;
                            Time,
                            Depth       : Byte);


IMPLEMENTATION
USES
  Crt,
  LookUpTables;
{-----------------------------------------------------------------------------------------}

FUNCTION Flip(Sq : Int8) : Int8; INLINE;
BEGIN
  Flip := -1;
  IF Sq <> -1 THEN
    Flip := Sq XOR 56;
END;


FUNCTION WhatPiece(Sq : Int8) : PieceType; INLINE; {-- what piece is on this square }
VAR
  Rank,
  Fyle : Byte;
BEGIN
  Rank := Sq SHR 3;
  Fyle := Sq MOD 8;

  {-- move the nibble to the LSBits and mask it off }
  WhatPiece := ( Data[Ply].Board[Rank] SHR (4*Fyle) ) AND %1111;
END;


FUNCTION WhatFile(Sq : Int8): Byte; INLINE;
BEGIN
  WhatFile := Sq MOD 8;
END;


FUNCTION WhatRank(Sq : Int8) : Byte; INLINE;
BEGIN
  WhatRank := Sq SHR 3;
END;


FUNCTION SameRank(FromSq, ToSq : Int8) : Boolean; INLINE;
BEGIN
  SameRank := (FromSq SHR 3) = (ToSq SHR 3);
END;


FUNCTION SameFile(FromSq, ToSq : Int8) : Boolean; INLINE;
BEGIN
  SameFile := Abs(ToSq - FromSq) MOD 8 = 0;
END;


FUNCTION SameDiagonal(FromSq, ToSq : Int8) : Boolean;
VAR
  FromRank,
  FromFile,
  ToRank,
  ToFile : Int8;
BEGIN
  IF (FromSq = -1) OR (ToSq = -1) THEN  {-- cause fromsq may be no piece }
    Exit(False);

  FromRank := FromSq SHR 3;
  FromFile := FromSq MOD 8;
  ToRank   := ToSq   SHR 3;
  ToFile   := ToSq   MOD 8;

  SameDiagonal := Abs(FromRank - ToRank) = Abs(FromFile - ToFile);
END;


FUNCTION StrToToSq(Move : ShortStr) : Byte; INLINE;
VAR
  Rank,
  Fyle : Byte;
BEGIN
  Fyle := Ord(Move[3]) - 65;
  Rank := Ord(Move[4]) - 49;
  StrToToSq := (Rank SHL 3) + Fyle;
END;


FUNCTION StrToFromSq(Move : ShortStr) : Byte; INLINE;
VAR
  Rank,
  Fyle : Byte;
BEGIN
  Fyle := Ord(Move[1]) - 65;
  Rank := Ord(Move[2]) - 49;
  StrToFromSq := (Rank SHL 3) + Fyle;
END;


FUNCTION SquaresToStr(FromSq, ToSq : Int8) : ShortStr; INLINE;
BEGIN
  SquaresToStr := SqToStr[FromSq] + SqToStr[ToSq];
END;


FUNCTION ClearVector(FromSq, ToSq : Int8) : Boolean;
{-- is there any pieces between fromsq and tosq...used for queen/rook/bishop moves }
VAR
  i,
  Step,
  Delta : Int16;
BEGIN
  IF FromSq = ToSq THEN   {-- this can happen when called from Qsearch }
    Exit(false);

  Assert( (FromSq >= 0) AND (ToSq >= 0), 'error in ClearVector...sq = 0');

  Delta := Abs(ToSq - FromSq);

  IF (Delta MOD 9) = 0 THEN
    Step := 9
  ELSE
    IF (Delta MOD 8) = 0 THEN
      Step := 8
    ELSE
      IF (Delta MOD 7) = 0 THEN
        BEGIN
          IF SameRank(FromSq, ToSq) THEN  {-- rooks and queens }
            Step := 1
          ELSE
            Step := 7;                    {-- diagonal backslash }
        END
      ELSE
        Step := 1;

  IF ToSq < FromSq THEN
    Step := -Step;

  i := FromSq;
  REPEAT
    Inc(i, Step);
  UNTIL (WhatPiece(i) <> Empty) OR (i = ToSq);

  ClearVector := (i = ToSq);    {-- got there without bumping into a piece }
END;


FUNCTION Attacked(MySq : Int8; Enemy : Int8) : Boolean;
{-- is MySq square attacked }
VAR
  BlackIsEnemy : Boolean;
  i,
  FromSq : Int8;
BEGIN
  Attacked := False;  {-- unless proven otherwise }

  BlackIsEnemy := Enemy = Bmove;

  WITH Data[Ply] DO
    BEGIN
      IF BlackIsEnemy THEN
        BEGIN
          {-- pawn attacks }
          FOR i := 0 TO 7 DO
            IF (BpawnList[i] <> NoPiece) AND (Abs(BpawnList[i] - MySq) IN [7, 9]) THEN
              Exit(True);

          FOR i := 0 TO 1 DO
            BEGIN
              {-- Knight attacks }
              FromSq := BpieceList[i, Bnight];
              IF FromSq <> NoPiece THEN
                IF Abs(WhatFile(BpieceList[i, Bnight]) - WhatFile(MySq)) < 3 THEN  {-- wrap around issue }
                  IF Abs(FromSq - MySq) IN [6, 10, 15, 17] THEN
                    Exit(True);

              {-- Bishops attacks }
              FromSq := BpieceList[i, Bbishop];
              IF FromSq <> NoPiece THEN
                IF SameDiagonal(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN
                  Exit(True);

              {-- rook attacks }
              FromSq := BpieceList[i, Brook];
              IF FromSq <> NoPiece THEN
                BEGIN
                  IF SameFile(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN
                    Exit(True);
                  IF SameRank(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN
                    Exit(True);
                END;

              {-- queen attacks }
              FromSq := BpieceList[i, Bqueen];
              IF FromSq <> NoPiece THEN
                BEGIN
                  IF SameFile(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN     {-- vertical }
                    Exit(True);
                  IF SameRank(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN     {-- horizontal }
                    Exit(True);
                  IF SameDiagonal(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN {-- diagonal }
                    Exit(True);
                END;

              {-- king attacks }
              IF BpieceList[0, Bking] <> NoPiece THEN
                IF Abs(WhatFile(BpieceList[0, Bking]) - WhatFile(MySq)) < 2 THEN  {-- stupid wrap around again }
                  IF Abs(WhatRank(BpieceList[0, Bking]) - WhatRank(MySq)) < 2 THEN
                    IF Abs(BpieceList[0, Bking] - MySq) IN [1, 9] THEN
                      Exit(True);

            END; {-- for }
        END  {-- if }
      ELSE
        BEGIN  {-- white is enemy }
          {-- pawn attacks }
          FOR i := 0 TO 7 DO
            IF (WpawnList[i] <> NoPiece) AND (Abs(WpawnList[i] - MySq) IN [7, 9]) THEN
              Exit(True);

          FOR i := 0 TO 1 DO
            BEGIN
              {-- Knight attacks }
              FromSq := WpieceList[i, Wnight];
              IF FromSq <> NoPiece THEN
                IF Abs(WhatFile(WpieceList[i, Wnight]) - WhatFile(MySq)) < 3 THEN  {-- wrap around issue }
                  IF Abs(FromSq - MySq) IN [6, 10, 15, 17] THEN
                    Exit(True);

              {-- Bishops attacks }
              FromSq := WpieceList[i, Wbishop];
              IF FromSq <> NoPiece THEN
                IF SameDiagonal(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN
                  Exit(True);

              {-- rook attacks }
              FromSq := WpieceList[i, Wrook];
              IF FromSq <> NoPiece THEN
                BEGIN
                  IF SameFile(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN
                    Exit(True);
                  IF SameRank(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN
                    Exit(True);
                END;

              {-- queen attacks }
              FromSq := WpieceList[i, Wqueen];
              IF FromSq <> NoPiece THEN
                BEGIN
                  IF SameFile(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN      {-- vertical }
                    Exit(True);
                  IF SameRank(FromSq, MySq)  AND ClearVector(FromSq, MySq) THEN     {-- horizontal }
                    Exit(True);
                  IF SameDiagonal(FromSq, MySq) AND ClearVector(FromSq, MySq) THEN  {-- diagonal }
                    Exit(True);
                END;

              {-- king attacks }
                IF Abs(WhatFile(WpieceList[0, Wking]) - WhatFile(MySq)) < 2 THEN  {-- stupid wrap around again }
                  IF Abs(WhatRank(WpieceList[0, Wking]) - WhatRank(MySq)) < 2 THEN
                    IF Abs(WpieceList[0, Wking] - MySq) IN [1, 9] THEN
                      Exit(True);
            END; {-- for }
        END;  {-- else }
    END;   {-- with }
END;


FUNCTION ValidMove(M : FullTTtype) : Boolean;
{-- poor mans function but TT clashes are rare and this should avoid crashing with an illegal move }
VAR
  FromSq,
  ToSq        : Int8;
BEGIN
  ValidMove := False;
  IF M.Move <> '' THEN
    BEGIN
      FromSq := StrToFromSq(M.Move);
      ToSq   := StrToToSq(M.Move);

      IF WhatPiece(FromSq) = M.Piece THEN
        BEGIN
          IF WhosMove = Wmove THEN
            ValidMove := (WhatPiece(ToSq) IN [Empty, Bpawn..Bqueen])
          ELSE
            ValidMove := (WhatPiece(ToSq) IN [Empty, Wpawn..Wqueen]);
        END;
    END;
END;


FUNCTION InCheck(Who : Int8) : Boolean;
BEGIN
  IF Who = Wmove THEN     {-- white }
    InCheck := Attacked(Data[Ply].WpieceList[0, Wking], Bmove) {-- gets Square of king and passes it on }
  ELSE
    InCheck := Attacked(Data[Ply].BpieceList[0, Bking], Wmove);
END;


FUNCTION MoveStrToMoveRec(Mstr : ShortStr) : MoveRec; INLINE;
VAR
  M : MoveRec;
  P : PieceType;
BEGIN
  IF Mstr = 'MATE' THEN
    Exit(NoMove);

  M.FromSq := StrToFromSq(Mstr);
  M.ToSq   := StrToToSq(Mstr);
  P := WhatPiece(M.FromSq);

  M.Capture   := WhatPiece(M.ToSq) <> Empty;
  M.Castle    := (P IN [Wking, Bking]) AND (Abs(M.FromSq - M.ToSq) = 2);
  M.Promotion := ( (WhosMove = 1)  AND (P = Wpawn) AND (M.ToSq > 55) ) OR
                 ( (WhosMove = -1) AND (P = Bpawn) AND (M.ToSq < 8) );

  MoveStrToMoveRec := M;
END;


FUNCTION EBF(Var Nodes : SaveNodesType; Len : Byte) : Real;
{-- average the last 3 plys ebf }
VAR
  Temp : ARRAY[0..2] OF Real;
  i    : 0..2;
BEGIN
  FOR i := 0 TO 2 DO
    IF Nodes[Len -1 -i] <> 0 THEN
      Temp[i] := Nodes[Len - i] / Nodes[Len - 1 -i];
  EBF := (Temp[0] + Temp[1] + Temp[2]) / 3;
END;


FUNCTION IsMoveCmd(S : Str4) : Boolean;
BEGIN
  IsMoveCmd := (S[1] IN ['A'..'H']) AND (S[2] IN ['1'..'8']) AND (S[3] IN ['A'..'H']) AND (S[4] IN ['1'..'8']);
END;


PROCEDURE ErrorMsg(S : String);
CONST
  x = 49;
  y = 14;
BEGIN
  GoToXY(x, y);
  Write(S, ': (Press any key)');
  REPEAT
  UNTIL KeyPressed;
  GoToXY(x, y);
  ClrEol;
END;


FUNCTION UpperCase(S : String) : String;
VAR
  i : Byte;
BEGIN
  FOR i := 1 TO Length(S) DO
    S[i] := UpCase(S[i]);
  UpperCase := S
END;


FUNCTION WhiteMaterial : Int16;
VAR
  Mat : Int16;
  i   : Byte;
BEGIN
  Mat := 0;
  WITH Data[Ply] DO
    BEGIN
      FOR i := 0 TO 7 DO
        IF WpawnList[i] <> NoPiece THEN
          Mat := Mat + PieceValue[Wpawn];

      FOR i := 0 TO 1 DO
        BEGIN
          IF WpieceList[i, Wnight] <> NoPiece THEN
            Mat := Mat + PieceValue[Wnight];

          IF WpieceList[i, Wbishop] <> NoPiece THEN
            Mat := Mat + PieceValue[Wbishop];

          IF WpieceList[i, Wrook] <> NoPiece THEN
            Mat := Mat + PieceValue[Wrook];

          IF WpieceList[i, Wqueen] <> NoPiece THEN
            Mat := Mat + PieceValue[Wqueen];
        END;  {-- for }
    END;  {-- with }
  WhiteMaterial := Mat;
END;


FUNCTION BlackMaterial : Int16;
VAR
  Mat : Int16;
  i   : Byte;
BEGIN
  Mat := 0;
  WITH Data[Ply] DO
    BEGIN
      FOR i := 0 TO 7 DO
        IF BpawnList[i] <> NoPiece THEN
          Mat := Mat + PieceValue[Bpawn];

      FOR i := 0 TO 1 DO
        BEGIN
          IF BpieceList[i, Bnight] <> NoPiece THEN
            Mat := Mat + PieceValue[Bnight];

          IF BpieceList[i, Bbishop] <> NoPiece THEN
            Mat := Mat + PieceValue[Bbishop];

          IF BpieceList[i, Brook] <> NoPiece THEN
            Mat := Mat + PieceValue[Brook];

          IF BpieceList[i, Bqueen] <> NoPiece THEN
            Mat := Mat + PieceValue[Bqueen];
        END;  {-- for }
    END;  {-- with }
  BlackMaterial := Mat;
END;


FUNCTION CalcMaterial : Int16;  {-- is there a better way }
BEGIN
  CalcMaterial := WhiteMaterial - BlackMaterial;
END;


PROCEDURE SwapSides; INLINE;
BEGIN
  WhosMove   := -1*WhosMove;
  HumansMove := NOT HumansMove;
END;


FUNCTION IsEndGame : Boolean;
BEGIN
  IsEndGame := ( (WhiteMaterial < 3000) AND (Data[Ply].WpieceList[0, Wqueen] = -1) ) OR
               ( (BlackMaterial < 3000) AND (Data[Ply].BpieceList[0, Bqueen] = -1) );

END;


PROCEDURE AddToHistory(TheMove : MoveRec; Depth : Int8);
VAR
  Piece : PieceType;
  Sq    : Int8;
BEGIN
  Piece := WhatPiece(TheMove.FromSq);
  Inc(HistTab[Piece, TheMove.ToSq], Depth * Depth);

  {-- now age the table }
  IF HistTab[Piece, TheMove.ToSq] > 99 THEN    {-- age the whole table }
    BEGIN
      FOR Sq := 0 TO 63 DO
        BEGIN
          FOR Piece := Wpawn TO Wking DO
            HistTab[Piece, Sq] := HistTab[Piece, Sq] SHR 1;
          FOR Piece := Bpawn TO Bking DO
            HistTab[Piece, Sq] := HistTab[Piece, Sq] SHR 1;
        END;
    END;
END;


PROCEDURE WritePV(VAR PV : PvType; Len : Byte);
VAR
  i : Byte;
BEGIN
  Window(1, 1, 80, 25);

  GoToXY(2, 23);
  ClrEol;
  Write('PV: ');
  FOR i := 0 TO 13 DO
    IF i < Len THEN
      Write(PV[0, i] : 5);

  GoToXY(6, 24);
  ClrEol;
  FOR i := 14 TO (Len-1) DO
    IF i <= Len THEN
      Write(PV[0, i] : 5);
END;


PROCEDURE SetTime(S : Str4);
BEGIN
  Delete(S, 1, 1);
  MoveTime := StrToInt(S);
  FixedDepth := 255;
END;


PROCEDURE SetDepth(S : Str4);
BEGIN
  Delete(S, 1, 1);
  FixedDepth := StrToInt(S);
  MoveTime := 65000;
END;


PROCEDURE WriteInfoToScreen(Nodes,
                            Qnodes,
                            BetaCuts,
                            NullCuts,
                            TThits,
                            ExactTThits : Word32;
                            PVscore     : Int16;
                            PV          : PvType;
                            EBF         : Real;
                            Time,
                            Depth       : Byte);
{-- writes some useful info from the search for debugging and analysing }
BEGIN
  GoToXY(1, 1);
  IF EndGame THEN
    Write('*');

  Window(49, 2, 79, 12);
  ClrScr;

  Writeln('Nodes:' ,  Nodes  : 16);
  Write  ('QNodes:',  QNodes : 15);

  IF Nodes > 0 THEN
    Writeln(Round(100*QNodes/(Nodes+Qnodes)) : 4, '%');
  IF Time > 0 THEN
    Writeln('NPS:', (Nodes+Qnodes) DIV Time : 18)
  ELSE
    Writeln('NPS:', 0 : 18);

  Writeln('EBF:',       EBF : 18 : 1);   {-- branching factor }
  Writeln('Beta Cuts:', BetaCuts : 12);
  Writeln('Null Cuts:', NullCuts : 12);
  Writeln('TT Hits:',   ExactTThits : 6, '/', TThits : 7);
  Writeln('Time Used:', Time : 12);
  Writeln('Depth:',     Depth : 16);

  IF (Abs(PVscore) > 9980) THEN   {-- a mate in x moves }
    IF Abs(PVscore) = 9999 THEN
      Write('***** Check Mate *****')
    ELSE
      Write('Mate in:', (10000 - Abs(PVscore)) : 14)
  ELSE
    Write('Score:', WhosMove*PVscore : 16);

  WritePV(PV, Depth);
{  Write(' ', PVfromTT(MaxDepth));}


  if isendgame then
  begin
    gotoxy(1,1);
    write('EG');
  end;
END;




BEGIN
END.
