PROGRAM LtChess3;
USES
  Crt,
  Globals,
  LineInfo,
  Board,
  Search,
  Utils,
  Move,
  Opening,
  Debugging,
  Clocks;


VAR
  MyMove : MoveRec;
  Quit   : Boolean;


FUNCTION CalcMove : MoveRec;
{-- returns a move from the library or else searches for a move }
VAR
  LibMove  : ShortStr;
BEGIN
  IF InLibrary THEN
    BEGIN
      LibMove    := GetLibraryMove(OpeningStr);      {-- pass the opening string so far, to look for a match }
      OpeningStr := OpeningStr + LibMove;
      InLibrary  := LibMove <> 'A1A1';               {-- A1A1 is a sentinal. hopefully to avoid a crash }

      IF InLibrary THEN
        CalcMove := MoveStrToMoveRec(LibMove);
    END;

  IF NOT InLibrary THEN
    CalcMove := ComputerMove;
END;


PROCEDURE ReadLnCmd(VAR S : ShortStr);
{-- gets user command }
BEGIN
  GoToXY(49, 13);
  Write('Move: ');
  ClrEol;
  GoToXY(55, 13);
  Readln(S);
  S := UpperCase(S);
END;

{---------------------------------------------------------------------------------------}
BEGIN
  ClrScr;

  SetUpStartPosition;
  DisplayTheBoard;

  Quit := False;
  REPEAT  {-- game loop }
    CmdStr := 'A1A1';

    IF HumansMove THEN
      BEGIN
        ReadLnCmd(CmdStr);                       {-- get a move or command }

        IF IsMoveCmd(CmdStr) THEN                {-- a move }
          BEGIN
            IF InLibrary THEN                    {-- update the opening string }
              OpeningStr := OpeningStr + CmdStr;

            IF IsLegalMove(CmdStr) THEN
              MakeMove(MoveStrToMoveRec(CmdStr), False)  {-- convert it to a full move description and make the move }
            ELSE
              ErrorMsg('Illegal Move');
          END;
      END
    ELSE
      BEGIN
        MyMove := CalcMove;
        IF (MyMove.ToSq <> -1) THEN
          BEGIN
            MakeMove(MyMove, False);
            GoToXY(49, 14);
            Write('Move: ', SquaresToStr(MyMove.FromSq, MyMove.ToSq));
          END
        ELSE
          HumansMove := Not HumansMove;
      END;

    IF IsMoveCmd(CmdStr) THEN
      AddToTakeBackStack                    {-- takeback stack allows 16 takeback moves }
    ELSE
      CASE CmdStr[1] OF                     {-- process the command }
        'N' : SetUpStartPosition;           {-- start a new game }
        'P' : HumansMove := NOT HumansMove; {-- force computer move }
        'B' : TakeBackMove;                 {-- take back last move }
        'X' : SetDepth(CmdStr);             {-- fixed depth to search }
        'T' : SetTime(CmdStr);              {-- fixed time to search }
        'R' : ReadFen;                      {-- read a saved board from disk }
        'W' : WriteFen;                     {-- save a board position to disk }
        'S' : SetUpBoard;                   {-- set a position on the board }
        'E' : DisplayEval;                  {-- write some evaluation info to screen, when testing }
        'D' : Debug := False;               {-- true to run a debug routine }
        'Q' : Quit  := True;
        ELSE
          ErrorMsg('Illegal Command');
      END; {-- case}

    DisplayTheBoard;

  UNTIL Quit;
END.

