UNIT TT;

INTERFACE
USES
  Globals;

{$BITPACKING ON}
{$PACKENUM 1}
{$PACKRECORDS 1}

CONST
  NumTTentries  = 50331653;   {-- a prime number }

TYPE
  NodeType   = (None, Exact, FailLow, FailHigh);

  {-- the TT type that the program uses }
  FullTTtype = RECORD
                 Key   : Word64;     {-- of the board position }
                 Score : Int16;      {-- the score found in this position }
                 Node  : NodeType;   {-- what type of node was it }
                 Depth : Byte;       {-- how far to the leaf...the longer the better }
                 Move  : ShortStr;   {-- the best move }
                 Piece : PieceType;  {-- moving piece }
               END;

  {-- the packed tt record that the hash table uses to save space and allow more entries }
  InternTTtype = BITPACKED RECORD
                             Key        : Word64;     {-- of the board position }
                             Score      : Int16;      {-- the score found in this position }
                             Node_Depth : Byte;       {-- compressed into a byte }
                             Move_Piece : Word16;     {-- the best move and the piece }
                           END;


VAR
  TransTable : ARRAY[0..NumTTentries] OF InternTTtype;    {-- the transposition table }
  TTrecord   : FullTTtype;                                {-- used by the external program }

PROCEDURE ZeroTTs;

PROCEDURE AddToTT(Key   : Word64;       {-- of the board position }
                  Score : Int16;        {-- the score found in this position }
                  Node  : NodeType;     {-- what type of node was it }
                  Depth : Byte;         {-- draft...how far to the leaf...the longer the better }
                  Move  : ShortStr;     {-- move string }
                  Piece : PieceType);   {-- the best move }


PROCEDURE GetFromTT(VAR Entry : FullTTtype);


{-----------------------------------------------------------------------------}
IMPLEMENTATION
USES
  Zorbist,
  Utils,
  LookUpTables;


PROCEDURE ZeroTTs;
BEGIN
  FillChar(TransTable, SizeOf(TransTable), 0);
END;


PROCEDURE AddToTT(Key   : Word64;     {-- of the board position }
                  Score : Int16;      {-- the score found in this position }
                  Node  : NodeType;   {-- what type of node was it }
                  Depth : Byte;       {-- draft....how far to the leaf...the longer the better }
                  Move  : ShortStr;
                  Piece : PieceType);
VAR
  Slot      : Word64;
  MovePiece : Word16;
BEGIN
  IF WhosMove = -1 THEN                      {-- account for whos move it is }
    Key := NOT Key;

  Slot := Key MOD NumTTentries;              {-- find a slot }

  IF Depth >= (TransTable[Slot].Node_Depth AND %1111) THEN   {-- better infomation }
    BEGIN
      {-- pack the fromsq, tosq and piece into 16bits }
      IF Move <> '' THEN
        MovePiece := StrToFromSq(Move) OR (StrToToSq(Move) SHL 6) OR (Piece SHL 12)
      ELSE
        MovePiece := 0;  {-- cause a fail low does not save a move }

      TransTable[Slot].Key        := Key;
      TransTable[Slot].Score      := Score;
      TransTable[Slot].Node_Depth := Ord(Byte(Node) SHL 4) OR Depth;
      TransTable[Slot].Move_Piece := MovePiece;
    END;
END;


PROCEDURE GetFromTT(VAR Entry : FullTTtype);
{-- returns the node info in "entry" }
VAR
  Slot : Word64;
BEGIN
  IF WhosMove = -1 THEN
    Entry.Key := NOT Entry.Key;           {-- Different key depending on whos move it is }

  Slot := Entry.Key MOD NumTTentries;

  WITH Entry DO
    IF Key = TransTable[Slot].Key THEN    {-- a correct match }
      BEGIN
        Score := TransTable[Slot].Score;
        Node  := NodeType(TransTable[Slot].Node_Depth SHR 4);
        Depth := TransTable[Slot].Node_Depth AND %1111;
        Move  := SqToStr[Byte(TransTable[Slot].Move_Piece AND %111111)] +
                 SqToStr[Byte(TransTable[Slot].Move_Piece SHR 6) AND %111111];
        Piece := Byte(TransTable[Slot].Move_Piece SHR 12);
      END
    ELSE
      BEGIN
        Score := 0;
        Node  := None;
        Depth := 0;
        Move  := '';
        Piece := Empty;
      END;
END;



BEGIN

END.
