UNIT Zorbist;

INTERFACE
USES
  Globals;


TYPE   {-- a random number for every piece on every square }
  SqType      = Int8;
  ZorbBrdType = ARRAY[PieceType, SqType] OF Word64;

VAR
  ZorbistBrd : ZorbBrdType;   {-- the random numbers for each piece/colour/square }


PROCEDURE InitZorbistBoard;   {-- values for each piece/square }
FUNCTION  CalcBrdKey : Word64;

{-----------------------------------------------------------------------------}
IMPLEMENTATION



PROCEDURE InitZorbistBoard;
{-- calculate a random 64 bit number for each piece on each square }
VAR
  Sq    : SqType;
  Piece : PieceType;
BEGIN
  FillChar(ZorbistBrd, SizeOf(ZorbistBrd), 0);

  {-- calculate the entire table }
  {-- NoPiece has a zero value so will XOR to itself }
  Randomize;
  FOR Sq := 0 TO 63 DO              {-- for each piece }
    FOR Piece := Wpawn TO Bking DO  {-- with each piece on it }
      IF Piece IN [Wpawn..Wking, Bpawn..Bking] THEN
        ZorbistBrd[Piece, Sq] := (Word64(Random($FFFFFFFF)) SHL 32) OR
                                    Word64(Random($FFFFFFFF));
  Data[Ply].HashKey := CalcBrdKey;
END;


FUNCTION CalcBrdKey : Word64;
{-- calculate the zorist Key for the entire board }
VAR
  Value : Word64;
  i,
  Sq,
  Piece : Int8;
BEGIN
  Value := 0;
  {-- the pieces first }
  FOR i := 0 TO 1 DO
    BEGIN
      FOR Piece := Wnight TO Wking DO
        BEGIN
          Sq := Data[Ply].WpieceList[i, Piece];
          IF Sq <> -1 THEN
            Value := Value XOR ZorbistBrd[Piece, Sq];
        END;
      FOR Piece := Bnight TO Bking DO
        BEGIN
          Sq := Data[Ply].BpieceList[i, Piece];
          IF Sq <> -1 THEN
            Value := Value XOR ZorbistBrd[Piece, Sq];
        END;
      END;
  {-- then the pawns }
  FOR i := 0 TO 7 DO
    BEGIN
      Sq := Data[Ply].WpawnList[i];
      IF Sq <> -1 THEN
        Value := Value XOR ZorbistBrd[Wpawn, Sq];

      Sq := Data[Ply].BpawnList[i];
      IF Sq <> -1 THEN
        Value := Value XOR ZorbistBrd[Bpawn, Sq];
    END;

  CalcBrdKey := Value;
END;



{-----------------------------------------------------------------------------}
BEGIN
END.


